#include <stdio.h>
#include <string.h>
#include <assert.h>

#define N 20

static void
Aprod(int *n, double x[], double y[])
/* computes y := Ax */
{
    int i;

    assert(*n == N);
#if 0
    memcpy(y, x, N * sizeof(y[0]));
#else
    for (i = 0; i < N; i++)
	y[i] = (i==0 ? 0.0 : -x[i-1]) + 2. * x[i] + (i==N-1 ? 0.0 : -x[i+1]);
#endif
}

static void
Msolve(int *n, double x[], double y[])
/* solves My = x */
{
    int i;

    assert(*n == N);
#if 0
    memcpy(y, x, N * sizeof(y[0]));
#else
    for (i = 0; i < N; i++)
	y[i] = x[i] * 0.5;
#endif
}

static double
dot_product(double x[], double y[])
{
    int i;
    double d = x[0] * y[0];

    for (i = 1; i < N; i++)
	d += x[i] * y[i];

    return d;
}

static double rtol = 1e-3, atol = 0.0, btol = 1e-3;
static double bnorm, rnorm0;

static int
check_res(int *n, double *x, double *rnorm, int *itn)
/* Note: returns !0 will stop MINRES iterations */
{
    assert(*n == N);
    if (*itn == 1) {
	bnorm = 1.0;
	rnorm0 = *rnorm;
	if (rnorm0 <= 0.0)
	    rnorm0 = 1.0;
    }
    printf("it. number %d, res_a = %0.4e, res_b = %0.4e, res_r = %0.4e\n",
		*itn, *rnorm, *rnorm / bnorm, *rnorm / rnorm0);

    return (*rnorm<=atol || *rnorm<=btol*bnorm || *rnorm<=rtol*rnorm0) ? 1 : 0;
}

int
main(int argc, char *argv[])
{
    int n = N;
    extern void minreswrapper_( int *n,
				void (*Aprod)(int*, double*, double*),
				void (*Msolve)(int*, double*, double*),
				double *b, double *shift,
				int *checkA, int *precon,
				double *x, int *itnlim, int *nout, double *rtol,
				int *istop, int *itn,
				double *Anorm, double *Acond,
				double *rnorm, double *Arnorm, double *ynorm,
				double (*dot_product)(double*, double*),
				int (*check_res)(int*, double*, double*, int*) );
    /* input variables */
    double b[N], shift = 0.0, rtol = 0.0;
    int checkA = 0, precon = 1, itnlim = 100, nout = 0/*6*/;
    /* output variables */
    double x[N], Anorm, Acond, rnorm, Arnorm, ynorm;
    int istop, itn;

    memset(x, 0, sizeof(x));
    memset(b, 0, sizeof(b));
    b[0] = b[N - 1] = 1.0;
    minreswrapper_(&n, Aprod, Msolve, b, &shift, &checkA, &precon,
		   x, &itnlim, &nout, &rtol,
                   &istop, &itn, &Anorm, &Acond, &rnorm, &Arnorm, &ynorm,
                   dot_product, check_res);
#if 0
    printf("istop = %d, itn = %d, Anorm = %g, Acond = %g, rnorm = %g\n",
		istop, itn, Anorm, Acond, rnorm);
    for (n = 0; n < N; n++)
	printf("x[%d] = %0.15e\n", n, x[n]);
#endif
}
